//#define __SIM__

#ifdef __SIM__
#include <16F877.h>
#else
#include <16F690.h>
#endif

#device adc=8

#FUSES WDT                      //Watch Dog Timer
#FUSES HS                       //HS Oscilator

#ifdef __SIM__
#use i2c(Slave,Fast,sda=PIN_C4,scl=PIN_C3,restart_wdt,force_hw,address=0xAC)
#else
#FUSES NOPROTECT                //Code not protected from reading
#FUSES NOBROWNOUT               //No brownout reset
#FUSES NOMCLR                   //Master Clear pin used for I/O
#FUSES NOCPD                    //No EE protection
#FUSES NOPUT                    //No Power Up Timer
#FUSES NOIESO                   //Internal External Switch Over mode disabled
#FUSES NOFCMEN                  //Fail-safe clock monitor disabled
#use i2c(Slave,Fast,sda=PIN_B7,scl=PIN_B6,restart_wdt,force_hw,address=0xAC)
#endif

#use delay(clock=20000000)

//how big the buffer is .. ~78 is the largest I could make on 16f690
//33 is enough for lcd operations, everything else is pure bonus :)
#define ALCTD 75

BYTE incoming, state;
BYTE address;

BYTE buffer[ALCTD];
BYTE changed[ALCTD];


#int_SSP
void  SSP_isr(void) 
{
  state = i2c_isr_state();
  if(state == 1){  //First received byte is address
    incoming = i2c_read();
    address = incoming;
    if (address > ALCTD) address = 0;
  }
  if(state == 2){  //Second received byte is data
    incoming = i2c_read();
    buffer[address] = incoming;
    changed[address] = 1;
  }
  if(state == 0x80){  //Master is asking for data
    i2c_write (buffer[address]);
  }
}


// PIC     LCD
// C0      enable
// C1      rs
// C2      rw
// C3      NC
// C4      D4
// C5      D5
// C6      D6
// C7      D7


struct lcd_pin_map {                 // This structure is overlayed
           BOOLEAN enable;           // on to an I/O port to gain
           BOOLEAN rs;               // access to the LCD pins.
           BOOLEAN rw;               // The bits are allocated from
           BOOLEAN unused;           // low order up.  ENABLE will
           int     data : 4;         // be pin B0.
        } lcd;

#ifdef __SIM__
#locate lcd = getenv("sfr:PORTD")    // This puts the entire structure over the port
#define set_tris_lcd(x) set_tris_d(x)
#else
#locate lcd = getenv("sfr:PORTC")    // This puts the entire structure over the port
#define set_tris_lcd(x) set_tris_c(x)
#endif



#define lcd_type 2           // 0=5x7, 1=5x10, 2=2 lines
#define lcd_line_two 0x40    // LCD RAM address for the second line

BYTE const LCD_INIT_STRING[4] = {0x20 | (lcd_type << 2), 0xc, 1, 6};
                             // These bytes need to be sent to the LCD
                             // to start it up.

                             // The following are used for setting
                             // the I/O port direction register.

struct lcd_pin_map const LCD_WRITE = {0,0,0,0,0}; // For write mode all pins are out
struct lcd_pin_map const LCD_READ = {0,0,0,0,15}; // For read mode data pins are in

BYTE lcd_read_byte() {
      BYTE low,high;
      set_tris_lcd(LCD_READ);
      lcd.rw = 1;
      delay_cycles(1);
      lcd.enable = 1;
      delay_cycles(1);
      high = lcd.data;
      lcd.enable = 0;
      delay_cycles(1);
      lcd.enable = 1;
      delay_us(1);
      low = lcd.data;
      lcd.enable = 0;
      set_tris_lcd(LCD_WRITE);
      return( (high<<4) | low);
}

void lcd_send_nibble( BYTE n ) {
      lcd.data = n;
      delay_cycles(1);
      lcd.enable = 1;
      delay_us(2);
      lcd.enable = 0;
}


void lcd_send_byte( BYTE address, BYTE n ) {
      lcd.rs = 0;
      while ( bit_test(lcd_read_byte(),7) ) ;
      lcd.rs = address;
      delay_cycles(1);
      lcd.rw = 0;
      delay_cycles(1);
      lcd.enable = 0;
      lcd_send_nibble(n >> 4);
      lcd_send_nibble(n & 0xf);
}


void lcd_init() {
    BYTE i;
    set_tris_lcd(LCD_WRITE);
    lcd.rs = 0;
    lcd.rw = 0;
    lcd.enable = 0;
    delay_ms(15);
    for(i=1;i<=3;++i) {
       lcd_send_nibble(3);
       delay_ms(5);
    }
    lcd_send_nibble(2);
    for(i=0;i<=3;++i)
       lcd_send_byte(0,LCD_INIT_STRING[i]);
}

void main()
{
   BYTE i;

#ifdef __SIM__
#else
   setup_adc_ports(NO_ANALOGS|VSS_VDD);
   setup_adc(ADC_OFF);
   setup_timer_0(RTCC_INTERNAL|RTCC_DIV_1);
   setup_wdt(WDT_2304MS|WDT_DIV_16);
   setup_timer_1(T1_DISABLED);
#endif
   lcd_init();

   enable_interrupts(INT_SSP);
   enable_interrupts(GLOBAL);

   //init buffers
   for(i=0;i<32;i++){
     changed[i]=0;
     buffer[i]=0;
   }
  
  
  while (TRUE){
    #asm
    clrwdt
    #endasm

    //update 1st line
    for(i=0;i<16;i++){
      if (changed[i]){
        changed[i]=0;
        lcd_send_byte(0,0x80|i); //go to 1.row, i'th place
        lcd_send_byte(1,buffer[i]);        
      }
    }
    
    //update 2nd line
    for(i=16;i<32;i++){
      if (changed[i]){
        changed[i]=0;
        lcd_send_byte(0,0x80|(i+lcd_line_two-16)); //go to 2.row, i'th place
        lcd_send_byte(1,buffer[i]);        
      }
    }
    
    if (changed[32]){ // send controll code
      changed[32] = 0;
      lcd_send_byte(0,1); //clear screen    
    }
   
    for(i=33;i<ALCTD;i++){
     //do whatever .. controll the free oi pins on the uC...
    }
  
  }

}
